# encoding: utf-8
"""
Item Description
================
"""
__author__ = "Richard Smith"
__date__ = "27 May 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

import logging
from functools import lru_cache

# Typing imports
from typing import List, Optional

# Python imports
import yaml

# 3rd Party Imports
from directory_tree import DatasetNode

# Package imports
from asset_scanner.core.utils import dict_merge, load_description_files

LOGGER = logging.getLogger(__name__)


class ItemDescription:
    """
    Container to provide convenient access points into parts of
    the item description.
    """

    def __init__(self, description):
        self._description = description

    def __repr__(self):
        return yaml.dump(self._description)

    @property
    def defaults(self) -> dict:
        """Returns defaults"""
        return self._description.get("defaults", {})

    @property
    def overrides(self) -> Optional[dict]:
        """Returns overrides"""
        return self._description.get("overrides")

    @property
    def mappings(self) -> Optional[dict]:
        """Returns mappings"""
        return self._description.get("mappings")

    @property
    def allowed_facets(self) -> List:
        """Returns allowed facets"""
        return self.facet_extract_conf.get("allowed_facets", [])

    @property
    def extraction_methods(self) -> List[dict]:
        """Returns extraction methods"""
        return self.facet_extract_conf.get("extraction_methods", [])

    @property
    def aggregation_facets(self) -> List:
        """Returns aggregation facets"""
        return self.facet_extract_conf.get("aggregation_facets", [])

    @property
    def facet_extract_conf(self) -> dict:
        """Returns facets key"""
        return self._description.get("facets", {})

    @property
    def categories(self):
        """Returns categories"""
        return self._description.get("categories", [])

    @property
    def collection(self):
        """Returns collection"""
        return self._description.get("collection", {})

    @property
    def search_facets(self) -> List:
        """Returns extra top level facets"""
        return self.facet_extract_conf.get("search_facets", [])

    @property
    def templates(self):
        """Return the templates section of the description"""
        return self._description.get("templates", {})

    @property
    def title_template(self):
        """Returns the template for the title"""
        return self.templates.get("title_template")

    @property
    def description_template(self):
        """Returns the template for the title"""
        return self.templates.get("description_template")


class ItemDescriptions:
    """
    Holds references to all the description files and handles loading, merging
    and returning an :py:obj:`ItemDescription`
    """

    def __init__(self, root_path: str):
        """

        :param root_path: Path to the root of the yaml store
        """

        self.tree = DatasetNode()

        self._build_tree(root_path)

    def _build_tree(self, root_path: str) -> None:
        """
        Loads the yaml files from the root path and builds the dataset tree
        with references to the yaml files.

        :param root_path: Path at the top of the yaml file tree
        """

        files = load_description_files(root_path)

        if not files:
            LOGGER.error(
                "No description files found. "
                "Check the path in your configuration. Exiting..."
            )
            exit()

        for file in files:
            with open(file) as reader:
                data = yaml.safe_load(reader)

                for dataset in data.get("datasets", []):
                    # Strip trailing slash.
                    # Needed to make sure tree search works
                    dataset = dataset.rstrip("/")

                    self.tree.add_child(dataset, description_file=file.as_posix())

    def get_description(self, filepath: str) -> ItemDescription:
        """
        Get the merged description for the given file path.
        This gets all the description files along the path
        and merges them from top down so that more generic
        descriptions are overridden.
        e.g.

        files describing ``/badc`` will be overridden by files
        which describe ``/badc/faam/data``

        Dict values are overridden by more specific files and
        arrays are appended to, with duplicates ignored.

        .. note::
            For remote filepaths (e.g. https://... or gs://) a ``/``
            character will be pre-pended. This is to enable the lookup
            to pass as the root node of the tree is ``/``.

        :param filepath: Path for which to retrieve the description
        """

        if not filepath[0] == "/":
            filepath = f"/{filepath}"

        nodes = self.tree.search_all(filepath)
        description_files = [node.description_file for node in nodes]

        config_description = self.load_config(*description_files)

        return ItemDescription(config_description)

    @lru_cache(100)
    def load_config(self, *args: str) -> dict:
        """

        :param args: each arg is a filepath to a description file
        :return: Dictionary containing the merged properties of all the matching nodes
        """
        base_dict = {}
        for file in args:
            with open(file) as reader:
                base_dict = dict_merge(base_dict, yaml.safe_load(reader))

        return base_dict


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("root")
    args = parser.parse_args()

    descriptions = ItemDescriptions(args.root)

    description = descriptions.get_description(
        "/badc/faam/data/2005/b069-jan-05/core_processed/core_faam_20050105_r0_b069.nc"
    )

    print(description)



# encoding: utf-8
"""
The output plugins determine what happens at the end of the
extraction process.

You can configure more than one active plugin, if you wanted
to output the content to more than one place.

Output plugins are loaded as named entry points with the namespace:
``asset_scanner.output_plugins``
"""
__author__ = "Richard Smith"
__date__ = "08 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "08 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

import argparse
import logging
from pydoc import locate

import pkg_resources
import yaml

from asset_scanner.core import BaseExtractor
from asset_scanner.core.exceptions import NoPluginsError
from asset_scanner.core.utils import load_plugins


def command_args():
    """
    Sets the command line arguments and handles their parsing
    :return: command line options
    """
    parser = argparse.ArgumentParser(description="Run the asset scanner as configured")
    parser.add_argument("conf", help="Path to a yaml configuration file")
    args = parser.parse_args()

    return args


def setup_logging(conf):
    config = conf.get("logging", {})
    if not config or (config and not config.get("format")):
        config["format"] = "%(asctime)s %(name)s %(levelname)s %(message)s"

    logging.basicConfig(**config)


def load_config(path):
    with open(path) as reader:
        conf = yaml.safe_load(reader)
    return conf


def load_extractor(conf: dict) -> BaseExtractor:
    """
    Load the extractor.

    Looks for extractor defined in the configuration in preference
    and falls back to the first defined entry point at ``asset_scanner.extractors``

    :param conf: Configuration dict
    :return: Extractor
    """

    extractor = None

    if conf.get("extractor"):
        extractor = locate(conf["extractor"])
        if not extractor:
            raise ImportError(
                f'Unable to find {conf["extractor"]}. Check that it is installed.'
            )

    if not extractor:
        for entry_point in pkg_resources.iter_entry_points("asset_scanner.extractors"):
            extractor = entry_point.load()

            # Only load the first one
            break

    if not extractor:
        raise NoPluginsError("No extraction plugins have been loaded")

    return extractor(conf)


def main():
    args = command_args()

    conf = load_config(args.conf)

    setup_logging(conf)

    extractor = load_extractor(conf)

    input_plugins = load_plugins(conf, "asset_scanner.input_plugins", "inputs")

    for input in input_plugins:
        input.run(extractor)


if __name__ == "__main__":
    main()



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "23 Sep 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

import enum


class StorageType(enum.Enum):
    """Enum values for storage classes"""

    POSIX = "POSIX"
    OBJECT_STORE = "OBJECT_STORE"
    TAPE = "TAPE"



# encoding: utf-8
"""
The input plugins determine the source of the list. Files are processed
atomically and the input plugins provide this atomic action.

You can configure more than one input plugin, if you wanted
to input the content from more than one place.

Input plugins are loaded as named entry points with the namespace:
``asset_scanner.input_plugins``

.. warning::
    Blocking input plugins will prevent others from being run. They are run
    sequentially. For example, with the `file system input`_ plugin, you
    could configure several to scan multiple directories but the rabbit plugin
    creates a listening connection which would block any other inputs.
"""
__author__ = "Richard Smith"
__date__ = "08 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"



# encoding: utf-8
"""
Base Extractor
--------------

This module provides the base class for all derived extractors.

"""
__author__ = "Richard Smith"
__date__ = "08 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

import re
from abc import ABC, abstractmethod
from typing import List

from asset_scanner.types.source_media import StorageType

from .handler_picker import HandlerPicker
from .item_describer import ItemDescription, ItemDescriptions
from .utils import load_plugins


class BaseExtractor(ABC):
    """
    Base class to define an interface for other extractor classes

    Attributes:

        PROCESSOR_ENTRY_POINT:
            Defines the entry point to look for in the setup.py for the
            downstream package. This is used by the ``asset_scanner`` command
            to load the extractor if the extractor is not specifically defined
            in the the configuration file.
    """

    PROCESSOR_ENTRY_POINT = None

    def __init__(self, conf: dict):
        self.conf = conf
        self.processors = self.load_processors()
        self.output_plugins = self.load_output_plugins()
        self.item_descriptions = (
            ItemDescriptions(conf["item_descriptions"]["root_directory"])
            if "item_descriptions" in conf
            else None
        )

        self.load_processors()

    @staticmethod
    def _get_category(string, label, regex):
        """

        :param string:
        :param label:
        :param regex:
        :return:

        """

        m = re.search(regex, string)

        if not m:
            label = None

        return label

    def get_categories(
        self, filepath: str, source_media: StorageType, description: ItemDescription
    ) -> List:
        """
        Get asset category labels

        :param filepath: Asset path
        :param source_media: Source media class
        :param description: ItemDescription for asset
        :return:

        """
        categories = set()

        for conf in description.categories:
            label = self._get_category(filepath, **conf)
            if label:
                categories.add(label)

        return list(categories) or ["data"]

    def load_processors(self, entrypoint: str = None) -> HandlerPicker:
        return HandlerPicker(entrypoint or self.PROCESSOR_ENTRY_POINT)

    def load_output_plugins(self) -> List:
        return load_plugins(self.conf, "asset_scanner.output_plugins", "outputs")

    @abstractmethod
    def process_file(
        self, filepath: str, source_media: StorageType = StorageType.POSIX, **kwargs
    ) -> None:
        pass

    def output(
        self,
        filepath: str,
        source_media: StorageType,
        data: dict,
        namespace: str = None,
    ) -> None:
        for backend in self.output_plugins:
            if not backend.namespace or backend.namespace == namespace:
                backend.export(data)



# encoding: utf-8
"""
File System Input
-----------------

Takes a path and will scan the file system, submitting
each file to the asset extractor

**Plugin name:** ``file_system``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``path``
      - ``string``
      - ``REQUIRED`` The root path to scan
    * - ``kwargs``
      - ``dict``
      - Optional kwargs to pass to `os.walk <https://docs.python.org/3/library/os.html#os.walk>`_
    * - ``filters``
      - :ref:`PluginFilter <asset_scanner/plugin_filters:plugin filters>`
      - Optional filter plugins

Example Configuration:
    .. code-block:: yaml

        inputs:
            - name: file_system
              path: test_directory

"""
__author__ = "Richard Smith"
__date__ = "02 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"


import logging
import os
from datetime import datetime
from typing import TYPE_CHECKING

from tqdm import tqdm

from asset_scanner.types.source_media import StorageType

from .base import BaseInputPlugin

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from asset_scanner.core import BaseExtractor


class FileSystemInputPlugin(BaseInputPlugin):
    """
    Performs an os.walk to provide a stream of paths for procesing.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.root_path = kwargs["path"]
        self.kwargs = kwargs.get("kwargs", {})

    def run(self, extractor: "BaseExtractor"):
        total_files = 0
        start = datetime.now()
        for root, _, files in tqdm(os.walk(self.root_path, **self.kwargs)):
            for file in files:
                filename = os.path.abspath(os.path.join(root, file))

                if self.should_process(filename, StorageType.POSIX):
                    extractor.process_file(filename, StorageType.POSIX)
                    logger.debug(f"Input processing: {filename}")
                else:
                    logger.debug(f"Input skipping: {filename}")
                total_files += 1
        end = datetime.now()
        print(f"Processed {total_files} files from {self.root_path} in {end-start}")



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "02 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

from abc import ABC, abstractmethod

from asset_scanner.core import BaseExtractor
from asset_scanner.core.utils import load_plugins
from asset_scanner.types.source_media import StorageType


class BaseInputPlugin(ABC):
    def __init__(self, **kwargs):
        self.filters = None
        if kwargs.get("filters"):
            self.filters = load_plugins(
                kwargs, "asset_scanner.plugin_filters", "filters"
            )

    def should_process(self, filepath, source_media: StorageType) -> bool:
        """
        Should the path be sent for processing?

        Will run through any filter plugins. All plugins must pass for a True
        response. Any False will short circuit the logic and return False

        :param filepath: Filepath to test
        :param source_media: Source media

        :return: Bool, ``default: True``
        """
        if self.filters:
            return any((filter.run(filepath, source_media) for filter in self.filters))

        return True

    @abstractmethod
    def run(self, extractor: BaseExtractor):
        ...



# encoding: utf-8
"""
Object Store Input
------------------

Takes an endpoint url and optionally a bucket prefix and delimiter and will
scan the items at these points in the object store, submitting each to the
asset extractor

**Plugin name:** ``object_store``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``session_kwargs``
      - ``dict``
      - ``REQUIRED`` Dictionary containing the S3 access and secret keys
    * - ``endpoint_url``
      - ``string``
      - ``REQUIRED`` URL for S3 endpoint
    * - ``bucket``
      - ``string``
      - Bucket to be scanned. If ``None`` all buckets in endpoint will be scanned.
    * - ``prefix``
      - ``string``
      - Only items with prefix will be scanned
    * - ``delimiter``
      - ``string``
      - Group items after delimiter into one object

Example Configuration:
    .. code-block:: yaml

        inputs:
            - name: object_store
                endpoint_url: https://cedadev-o.s3-ext.jc.rl.ac.uk
                session_kwargs: {
                    aws_access_key_id: ACCESS_KEY,
                    aws_secret_access_key: SECRET_KEY
                }
                bucket: my_bucket
                prefix: directory_or_file
                delimiter: .zarr/

"""
__author__ = "Rhys Evans"
__date__ = "02 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "rhys.r.evans@stfc.ac.uk"


import logging
from typing import TYPE_CHECKING

import boto3

from .base import BaseInputPlugin

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from asset_scanner.core import BaseExtractor


class ObjectStoreInputPlugin(BaseInputPlugin):
    """ """

    def __init__(self, **kwargs):

        self.endpoint_url = kwargs.get("endpoint_url")
        session = boto3.session.Session(**kwargs["session_kwargs"])
        s3 = session.resource(
            "s3",
            endpoint_url=self.endpoint_url,
        )

        self.client = s3.meta.client
        self.buckets = (
            [s3.Bucket(kwargs["bucket"])] if "bucket" in kwargs else s3.buckets.all()
        )
        self.prefix = kwargs.get("prefix", "")
        self.delimiter = kwargs.get("delimiter", "")

    def run(self, extractor: "BaseExtractor"):

        for bucket in self.buckets:
            total_files = 0
            for obj in bucket.objects.filter(
                Prefix=self.prefix, Delimiter=self.delimiter
            ):
                extractor.process_file(
                    f"{self.endpoint_url}/{bucket.name}/{obj.key}",
                    "object_store",
                    client=self.client,
                )
                total_files += 1

            LOGGER.info(f"Processed {total_files} files from {bucket}")



# encoding: utf-8
"""
Standard Out
------------

An output backend which outputs the generated metadata to standard out.
Useful for testing and debugging.

**Plugin name:** ``standard_out``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``namespace``
      - ``str``
      - Can be used by downstream processors to separate outputs.

Example configuration:
    .. code-block:: yaml

        outputs:
            - name: standard_out
"""
__author__ = "Richard Smith"
__date__ = "01 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

from .base import OutputBackend


class StdoutOutputBackend(OutputBackend):
    """
    Simple print backend which can be used
    for testing and debugging.
    """

    def export(self, data: dict, **kwargs) -> None:
        """
        Print the received data.

        :param data: Data from extraction process
        :param kwargs: Not used
        """
        print(data)



# encoding: utf-8
"""
Thredds Input
-----------------

Uses a `Thredds Data Server <https://www.unidata.ucar.edu/software/tds/current/>`_
as a source.

**Plugin name:** ``thredds``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``uri``
      - ``string``
      - ``REQUIRED`` The URL to a Thredds Data Server.
    * - ``object_path_attr``
      - ``string``
      - ``REQUIRED`` The column header which contains the URI to
        the file object.
    * - ``catalog_kwargs``
      - ``dict``
      - Optional kwargs to pass to
        `siphon.catalog.TDSCatalog
        <https://unidata.github.io/siphon/latest/api/catalog.html#siphon.catalog.TDSCatalog>`_


Example Configuration:
    .. code-block:: yaml

        inputs:
            - name: thredds
              uri: test-url
              object_path_attr: access_urls.OPENDAP

"""
__author__ = "Mathieu Provencher"
__date__ = "3 Dec 2021"
__copyright__ = "Copyright 2021 Computer Research Institute of Montréal"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "mathieu.provencher@crim.ca"

# Python imports
import logging
from datetime import datetime
from typing import TYPE_CHECKING
from urllib.parse import urlparse

# Thirdparty imports
from siphon.catalog import CaseInsensitiveDict, TDSCatalog

# Framework imports
from asset_scanner.types.source_media import StorageType

# Package imports
from .base import BaseInputPlugin

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from asset_scanner.core import BaseExtractor


def walk_tds(cat: TDSCatalog, depth: int = 1):
    """
    Return a generator walking a THREDDS data catalog for datasets.

    :param cat: THREDDS catalog.
    :param depth: Maximum recursive depth. Setting 0 will return only datasets within the top-level catalog. If None,
      depth is set to 1000.
    """
    yield from cat.datasets.items()

    if depth is None:
        depth = 1000

    if depth > 0:
        for name, ref in cat.catalog_refs.items():
            child = ref.follow()
            yield from walk_tds(child, depth=depth - 1)


def get_sub_attr(obj: object, path: str):
    """
    Returns a child or sub-child attribute of a dict object.

    :param obj: Object
    :param path: 'attr1.attr2.etc'
    :return: obj.attr1.attr2.etc
    """
    attrs = path.split(".")

    for attr in attrs:
        if type(obj) == CaseInsensitiveDict:
            obj = obj[attr]
            continue

        obj = getattr(obj, attr)

    return obj


class ThreddsInputPlugin(BaseInputPlugin):
    """
    Process each dataset underneath a TDS catalog.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.uri = kwargs["uri"]
        self.object_attr = kwargs["object_path_attr"]

        self.thredds_kwargs = kwargs.get("catalog_kwargs", {})

    def open_catalog(self):
        """
        Open the Thredds catalog.

        :return: TDSCatalog
        """
        logger.info("Opening catalog")
        catalog = TDSCatalog(self.uri, **self.thredds_kwargs)

        return catalog

    def process_ds(self, ds, extractor: "BaseExtractor"):
        """
        Process a single dataset.

        :param ds: siphon.catalog.TDSCatalog.datasets
        :param extractor: BaseExtractor
        """
        filepath = get_sub_attr(ds, self.object_attr)
        parse_result = urlparse(filepath)

        # Set media type
        media_type = StorageType.OBJECT_STORE
        if not parse_result.netloc:
            media_type = StorageType.POSIX

        if self.should_process(filepath, media_type):
            extractor.process_file(filepath, media_type)
            logger.debug(f"Input processing: {filepath}")
        else:
            logger.debug(f"Input skipping: {filepath}")

    def run(self, extractor: "BaseExtractor"):
        """
        Plugin's entrypoint.

        :param extractor: BaseExtractor
        """
        total_files = 0
        start = datetime.now()

        catalog = self.open_catalog()

        for name, ds in walk_tds(catalog, depth=None):
            self.process_ds(ds, extractor)
            total_files += 1

        end = datetime.now()
        print(f"Processed {total_files} files from {self.uri} in {end - start}")



# encoding: utf-8
"""
Elasticsearch
-------------

An output backend which outputs the content generated to elasticsearch
using the Elasticsearch API

**Plugin name:** ``elasticsearch``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``connection_kwargs``
      - ``dict``
      - ``REQUIRED`` Connection kwargs passed to the `elasticsearch client  <https://elasticsearch-py.readthedocs.io/en/latest/api.html#elasticsearch>`_
    * - ``index.name``
      - ``str``
      - ``REQUIRED`` The index to output the content.
    * - ``index.mapping``
      - ``str``
      - Path to a yaml file which defines the mapping for the index
    * - ``namespace``
      - ``str``
      - Can be used by downstream processors to separate outputs to different indices or clusters

Example Configuration:
    .. code-block:: yaml

        outputs:
            - name: elasticsearch
              connection_kwargs:
                hosts: ['host1','host2']
              index:
                name: 'assets-2021-06-02'
              namespace: assets
"""
__author__ = "Richard Smith"
__date__ = "01 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

from typing import Dict

from elasticsearch import Elasticsearch

from asset_scanner.core.utils import Coordinates, load_yaml

from .base import OutputBackend


class ElasticsearchOutputBackend(OutputBackend):
    """
    Connects to an elasticsearch instance and exports the
    documents to elasticsearch.

    """

    CLEAN_METHODS = ["_format_bbox", "_format_temporal_extent"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        index_conf = kwargs["index"]

        self.es = Elasticsearch(**kwargs["connection_kwargs"])
        self.index_name = index_conf["name"]
        self.pipeline_conf = kwargs.get("pipeline", None)

        # Create the index, if it doesn't already exist
        if index_conf.get("mapping"):
            if not self.es.indices.exists(self.index_name):
                mapping = load_yaml(index_conf.get("mapping"))
                self.es.indices.create(self.index_name, body=mapping)

    @staticmethod
    def _format_bbox(data: Dict) -> Dict:
        """
        Convert WGS84 coordinates into GeoJSON and
        format for Elasticsearch. Replaces the bbox key.

        :param data: Input data dictionary
        """
        body = data["body"]

        if body.get("bbox"):
            bbox = body.pop("bbox")

            body["spatial"] = {
                "bbox": {
                    "type": "envelope",
                    "coordinates": Coordinates.from_wgs84(bbox).to_geojson(),
                }
            }

        return data

    @staticmethod
    def _format_temporal_extent(data: Dict) -> Dict:
        """
        Convert `extent object<https://github.com/radiantearth/stac-spec/blob/master/collection-spec/collection-spec.md#extent-object>_` for Elasticsearch.

        :param data: Input data dictionary
        """
        body = data["body"]

        if body.get("extent", {}).get("temporal"):
            temporal_extent = body["extent"].pop("temporal")

            if temporal_extent[0][0]:
                body["extent"]["temporal"] = {"gte": temporal_extent[0][0]}
            if temporal_extent[0][1]:
                temporal = body["extent"].get("temporal", {})
                temporal["lte"] = temporal_extent[0][1]
                body["extent"]["temporal"] = temporal

        return data

    def clean(self, data: Dict) -> Dict:
        """
        Condition the input dictionary for elasticsearch
        :param data: Input dictionary
        :returns: Dictionary produced as a result of the clean methods
        """

        for method in self.CLEAN_METHODS:
            m = getattr(self, method)
            data = m(data)

        return data

    def export(self, data, **kwargs):

        data = self.clean(data)

        index_kwargs = {
            "index": self.index_name,
            "id": data["id"],
            "body": {"doc": data["body"], "doc_as_upsert": True},
        }

        self.es.update(**index_kwargs)



# encoding: utf-8
"""
Exceptions
----------

"""
__author__ = "Richard Smith"
__date__ = "02 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"


class NoPluginsError(Exception):
    """
    Raised when no plugins have been loaded for a particular section.
    i.e. if no input plugins were successfully loaded, then this exception
    would be raised.
    """

    pass



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "08 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

import collections
import hashlib

# Python imports
import logging
from pathlib import Path

# Typing imports
from typing import Any, Dict, List, Optional, Union

# Third party imports
import yaml

from asset_scanner.core.exceptions import NoPluginsError
from asset_scanner.core.handler_picker import HandlerPicker

LOGGER = logging.getLogger(__name__)

NumType = Union[float, int]


class Coordinates:
    """
    Takes care of coordinate transformations
    """

    def __init__(self, minlon, maxlon, minlat, maxlat):
        self.minlon = minlon
        self.maxlon = maxlon
        self.minlat = minlat
        self.maxlat = maxlat

    @classmethod
    def from_geojson(cls, coordinates: List[List[NumType]]) -> "Coordinates":
        """
        GeoJSON formatted coordinates are in the form:

        [[minLon, maxLat],[maxLon, minLat]]

        :param coordinates: GeoJSON formatted coordinates from elasticsearch
        """

        minlon = coordinates[0][0]
        maxlon = coordinates[1][0]
        minlat = coordinates[1][1]
        maxlat = coordinates[0][1]

        return cls(minlon, maxlon, minlat, maxlat)

    @classmethod
    def from_wgs84(cls, coordinates: List) -> "Coordinates":
        """
        WGS84 formatted coordinates are in the form:

        [minLon, minLat, maxLon, maxLat]

        :param coordinates: WGS84 formatted coordinates
        """

        minlon = coordinates[0]
        maxlon = coordinates[2]
        minlat = coordinates[1]
        maxlat = coordinates[3]

        return cls(minlon, maxlon, minlat, maxlat)

    def to_wgs84(self) -> List[NumType]:
        """
        Exports the coordinates in WGS84 format

        [minLon, minLat, maxLon, maxLat]
        """

        return [self.minlon, self.minlat, self.maxlon, self.maxlat]

    def to_geojson(self) -> List[List[NumType]]:
        """
        Exports the coordinates in GeoJSON format

        [[minLon, maxLat],[maxLon, minLat]]
        """

        return [[self.minlon, self.maxlat], [self.maxlon, self.minlat]]


def load_plugins(conf: dict, entry_point: str, conf_section: str) -> List:
    """
    Load plugins from the entry points

    :param conf: Configuration dict
    :param entry_point: The name of the collection of entry points
    :param conf_section: The name for the section in the config file
    which applies to these plugins.

    Exceptions:
        NoPluginsError: Triggered if no plugins are successfully loaded

    :return: List of loaded plugins
    """

    loaded_plugins = []

    plugins = HandlerPicker(entry_point)

    for plugin_conf in conf[conf_section]:
        try:
            loaded_plugin = plugins.get_processor(**plugin_conf)
            loaded_plugins.append(loaded_plugin)
        except Exception:
            LOGGER.error(f'Failed to load plugin: {plugin_conf["name"]}', exc_info=True)

    if not loaded_plugins:
        raise NoPluginsError(f"No plugins were successfully loaded from {conf_section}")

    return loaded_plugins


def generate_id(path):
    return hashlib.md5(path.encode("utf-8")).hexdigest()


def load_yaml(path):
    with open(path) as reader:
        return yaml.safe_load(reader)


def dict_merge(*args, add_keys=True) -> dict:
    assert len(args) >= 2, "dict_merge requires at least two dicts to merge"

    # Make a copy of the root dict
    rtn_dct = args[0].copy()

    merge_dicts = args[1:]

    for merge_dct in merge_dicts:

        if add_keys is False:
            merge_dct = {
                key: merge_dct[key] for key in set(rtn_dct).intersection(set(merge_dct))
            }

        for k, v in merge_dct.items():

            # This is a new key. Add as is.
            if not rtn_dct.get(k):
                rtn_dct[k] = v

            # This is an existing key with mismatched types
            elif k in rtn_dct and not isinstance(v, type(rtn_dct[k])):
                raise TypeError(
                    f"Overlapping keys exist with different types: original is {type(rtn_dct[k])}, new value is {type(v)}"
                )

            # Recursive merge the next level
            elif isinstance(rtn_dct[k], dict) and isinstance(
                merge_dct[k], collections.abc.Mapping
            ):
                rtn_dct[k] = dict_merge(rtn_dct[k], merge_dct[k], add_keys=add_keys)

            # If the item is a list, append items avoiding duplictes
            elif isinstance(v, list):
                for list_value in v:
                    if list_value not in rtn_dct[k]:
                        rtn_dct[k].append(list_value)
            else:
                rtn_dct[k] = v

    return rtn_dct


def dot2dict(key: str, val: Any) -> Dict[str, Any]:
    """
    Convert a dot separated string into
    a dictionary construct. Can work with
    single layer or multi-layer strings.

    Recursively creates from bottom up.

    :param key: Key value. Can be single layer. e.g. ``properties`` or
                multi-level e.g. ``properties.start_time``
    :param val: The value associated with the key

    :return: dict
    """
    if not key:
        return val

    # Split on .
    parts = key.split(".")
    tail = parts.pop()

    # Create the new key
    key = ".".join(parts)

    val = {tail: val}
    return dot2dict(key, val)


def nested_get(key_list: List, input_dict: Dict) -> Optional[Any]:
    """
    Takes an iterable of keys and returns none if not found or the value
    :param key_list: List of keys to try against the dict
    :param input_dict: Dict to extract key from

    :return: Value found at the key location or None
    """

    last_key = key_list[-1]
    dict_nest = input_dict

    for key in key_list:
        if key != last_key:
            dict_nest = dict_nest.get(key, {})
        else:
            return dict_nest.get(key)


def load_description_files(path: str) -> List[str]:
    """
    Load the yaml description files recursively under the root path

    :param path: Root path for the description files
    """

    exts = [".yml", ".yaml"]
    return [p for p in Path(path).rglob("*") if p.suffix in exts]



# encoding: utf-8
"""
Path Regex Filter
-----------------

Takes a filepath and matches against a regex pattern. This can be used to pre-filter
paths for processing. Can use the ``exclude`` flag to flip the regex match.

**Plugin name:** ``path_regex``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``regex``
      - ``string``
      - ``REQUIRED`` Pattern to match.
    * - ``exclude``
      - ``bool``
      - Optional flag to either include or exclude the matched path. False (0) will send only paths that match to processing. True (1)
        will send all paths which do not match the regex for processing. ``DEFAULT: False``

Example Configuration:
    .. code-block:: yaml

        inputs:
            - name: file_system
              path: /badc/cmip5
              filters:
                - name: path_regex
                  regex: ^\/badc\/cmip[56]\/.*files
                  exclude: 1

"""
__author__ = "Richard Smith"
__date__ = "20 Sep 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

import re


class PathRegexFilter:
    def __init__(self, **kwargs):
        self.regex = kwargs["regex"]
        self.exclude = kwargs.get("exclude", 0)

    def run(self, filepath: str, source_media: str) -> bool:
        if self.exclude:
            return not bool(re.match(self.regex, filepath))
        else:
            return bool(re.match(self.regex, filepath))



# encoding: utf-8
"""

In some cases, it might be useful to be able to filter the response from the
plugins. E.g ignoring certain paths.

If more than one filter is specified, all tests must pass before processing
will continue.

Plugin filters are loaded as named entry points with the namespace:
``asset_scanner.plugin_filters``
"""
__author__ = "Richard Smith"
__date__ = "20 Sep 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "07 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

from abc import ABC, abstractmethod


class BaseProcessor(ABC):
    """
    Class to act as a base for all processors. Defines the basic method signature
    and ensure compliance by all subclasses.
    """

    def __init__(self, **kwargs):
        """
        Set the kwargs to generate instance attributes of the same name
        :param kwargs:
        """

        for k, v in kwargs.items():
            setattr(self, k, v)

    @abstractmethod
    def run(self, filepath: str, media_source: str = "POSIX", **kwargs) -> dict:
        """
        The action of running the processor and returning an output
        :param filepath: Path to object
        :param media_source: Media type for the target object
        :param kwargs: free kwargs passed to the processor.
        :return: dict
        """
        pass



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "01 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

from abc import ABC, abstractmethod


class OutputBackend(ABC):
    def __init__(self, namespace=None, **kwargs):
        self.namespace = namespace

    @abstractmethod
    def export(self, data, **kwargs):
        pass



# encoding: utf-8
"""
RabbitMQ Input
-----------------

Uses a `RabbitMQ Queue <https://www.rabbitmq.com/>`_ as a source for file objects.
At the moment, this expects a CEDA specifc message format.
TODO: Make the callback more flexible (open to collaboration)

**Plugin name:** ``rabbitmq``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``connection.host``
      - string
      - ``REQUIRED`` RabbitMQ server host
    * - ``connection.user``
      - string
      - ``REQUIRED`` Username
    * - ``connection.password``
      - string
      - ``REQUIRED`` password
    * - ``connection.vhost``
      - string
      - ``REQUIRED`` `Virtual host <https://www.rabbitmq.com/vhosts.html>`_
    * - ``connection.kwargs``
      - dict
      - connection parameter kwargs `pika.conneciton.ConnectionParameters
        <https://pika.readthedocs.io/en/stable/modules/parameters.html#connectionparameters>`_
    * - ``exchange.source_exchange``
      - dict
      - Dictionary describing the source exchange. `exchange`_
    * - ``exchange.dest_exchange``
      - dict
      - ``REQUIRED`` The final exchange. This is where the queues will be bound. `exchange`_
    * - ``queues``
      - ``list``
      - ``REQUIRED`` Queue parameters. `queues`_


exchange
^^^^^^^^

The source and dest exchange keys comprise:

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - name
      - string
      - ``REQUIRED`` Exchange name
    * - type
      - string
      - ``REQUIRED`` `Exchange type <https://medium.com/trendyol-tech/rabbitmq-exchange-types-d7e1f51ec825>`_

queues
^^^^^^

List of queue objects. Each queue object comprises:

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - name
      - string
      - ``REQUIRED`` Queue name
    * - kwargs
      - dict
      - kwargs passed to `pika.channel.queue_declare <https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.queue_declare>`_
    * - bind_kwargs
      - dict
      - kwargs passed to `pika.channel.queue_bind <https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.queue_bind>`_
    * - consume_kwargs
      - dict
      - kwargs passed to `pika.channel.Channel.basic_consume <https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.basic_consume>`_

Example Configuration:

    .. code-block:: yaml

        inputs:
            - name: rabbitmq
              connection:
                host: my-rabbit-server.co.uk
                user: user
                password: '*********'
                vhost: my_virtual_host
                kwargs:
                    heartbeat: 300
              exchange:
                source_exchange:
                    name: mysource-exchange
                    type: fanout
                destination_exchange:
                    name: mydest-exchange
                    type: fanout
              queues:
                - name:
                  kwargs:
                    durable: true
                  bind_kwargs:
                    routing_key: my.routing.key
                  consume_kwargs:
                    auto_ack: false


"""
__author__ = "Richard Smith"
__date__ = "29 Sep 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

# Python imports
import functools
import json
import logging
from collections import namedtuple

# Third-party imports
import pika

from asset_scanner.core import BaseExtractor
from asset_scanner.types.source_media import StorageType

from .base import BaseInputPlugin

LOGGER = logging.getLogger(__name__)

IngestMessage = namedtuple(
    "IngestMessage", ["datetime", "filepath", "action", "filesize", "message"]
)


class RabbitMQInputPlugin(BaseInputPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connection_conf = kwargs.get("connection", {})
        self.exchange_conf = kwargs.get("exchange", {})
        self.queues_conf = kwargs.get("queues", [])

    @staticmethod
    def decode_message(body: bytes) -> IngestMessage:
        """
        Takes the message and turns into a dictionary.
        String message format when split on :
            date_hour = split_line[0]
            min = split_line[1]
            sec = split_line[2]
            path = split_line[3]
            action = split_line[4]
            filesize = split_line[5]
            message = ":".join(split_line[6:])

        :param body: Message body, either a json string or text
        :return: IngestMessage
            {
                'datetime': ':'.join(split_line[:3]),
                'filepath': split_line[3],
                'action': split_line[4],
                'filesize': split_line[5],
                'message': ':'.join(split_line[6:])
            }

        """

        # Decode the byte string to utf-8
        body = body.decode("utf-8")

        try:
            msg = json.loads(body)
            return IngestMessage(**msg)

        except json.JSONDecodeError:
            # Assume the message is in the old format and split on :
            split_line = body.strip().split(":")

            msg = {
                "datetime": ":".join(split_line[:3]),
                "filepath": split_line[3],
                "action": split_line[4],
                "filesize": split_line[5],
                "message": ":".join(split_line[6:]),
            }

        return IngestMessage(**msg)

    def _connect(self, extractor) -> pika.channel.Channel:
        """
        Start Pika connection to server. This is run in each thread.

        :return: pika channel
        """

        # Get the username and password for rabbit
        rabbit_user = self.connection_conf.get("user")
        rabbit_password = self.connection_conf.get("password")

        # Get the server variables
        rabbit_server = self.connection_conf.get("host")
        rabbit_vhost = self.connection_conf.get("vhost")

        # Create the credentials object
        credentials = pika.PlainCredentials(rabbit_user, rabbit_password)

        # Start the rabbitMQ connection
        connection = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=rabbit_server,
                credentials=credentials,
                virtual_host=rabbit_vhost,
                **self.connection_conf.get("kwargs", {}),
            )
        )

        # Get the exchanges to bind
        src_exchange = self.exchange_conf.get("source_exchange")
        dest_exchange = self.exchange_conf.get("destination_exchange")

        # Create a new channel
        channel = connection.channel()

        # Declare relevant exchanges
        if src_exchange:
            channel.exchange_declare(
                exchange=src_exchange["name"], exchange_type=src_exchange["type"]
            )
        channel.exchange_declare(
            exchange=dest_exchange["name"], exchange_type=dest_exchange["type"]
        )

        # Bind source exchange to dest exchange
        if src_exchange:
            channel.exchange_bind(
                destination=dest_exchange["name"], source=src_exchange["name"]
            )

        # Declare queue and bind queue to the dest exchange
        for queue in self.queues_conf:
            declare_kwargs = queue.get("kwargs", {})
            bind_kwargs = queue.get("bind_kwargs", {})
            consume_kwargs = queue.get("consume_kwargs", {})

            channel.queue_declare(queue=queue["name"], **declare_kwargs)
            channel.queue_bind(
                exchange=dest_exchange["name"], queue=queue["name"], **bind_kwargs
            )

            # Set callback
            callback = functools.partial(
                self.callback, connection=connection, extractor=extractor
            )
            channel.basic_consume(
                queue=queue["name"], on_message_callback=callback, **consume_kwargs
            )

        return channel

    @staticmethod
    def _acknowledge_message(channel: pika.channel.Channel, delivery_tag: str):
        """
        Acknowledge message

        :param channel: Channel which message came from
        :param delivery_tag: Message id
        """

        LOGGER.debug(f"Acknowledging message: {delivery_tag}")
        if channel.is_open:
            channel.basic_ack(delivery_tag)

    def acknowledge_message(
        self,
        channel: pika.channel.Channel,
        delivery_tag: str,
        connection: pika.connection.Connection,
    ):
        """
        Acknowledge message and move onto the next. All of the required
        params come from the message callback params.

        :param channel: callback channel param
        :param delivery_tag: from the callback method param. eg. method.delivery_tag
        :param connection: connection object from the callback param
        """
        cb = functools.partial(self._acknowledge_message, channel, delivery_tag)
        connection.add_callback_threadsafe(cb)

    def callback(
        self,
        ch: pika.channel.Channel,
        method: pika.frame.Method,
        properties: pika.frame.Header,
        body: bytes,
        connection: pika.connection.Connection,
        extractor: BaseExtractor,
    ) -> None:

        # Get message
        try:
            message = self.decode_message(body)

        except IndexError:
            # Acknowledge message if the message is not compliant
            self.acknowledge_message(ch, method.delivery_tag, connection)
            return

        # Extract filename and storage class
        filename = message.filepath

        try:
            # TODO: How to get this from message? CEDA message.message is str
            # storage_class = message.message.get('storage_type', StorageType.POSIX)

            storage_class = StorageType.POSIX
            if isinstance(storage_class, str):
                storage_class = StorageType[storage_class]

        except KeyError:
            storage_class = StorageType.POSIX

        if self.should_process(filename, storage_class):
            extractor.process_file(filename, storage_class)

            LOGGER.debug(f"Input processing: {filename}")
            self.acknowledge_message(ch, method.delivery_tag, connection)
        else:
            LOGGER.debug(f"Input skipping: {filename}")

    def should_process(self, filepath, source_media: StorageType) -> bool:
        """
        Should the path be sent for processing?

        Will run through any filter plugins. All plugins must pass for a True
        response. Any False will short circuit the logic and return False

        :param filepath: Filepath to test
        :param source_media: Source media

        :return: Bool, ``default: True``
        """
        if self.filters:
            return any((filter.run(filepath, source_media) for filter in self.filters))

        return True

    def run(self, extractor: BaseExtractor):

        while True:
            channel = self._connect(extractor)

            try:
                LOGGER.info("READY")
                channel.start_consuming()

            except KeyboardInterrupt:
                channel.stop_consuming()
                break

            except pika.exceptions.StreamLostError as e:
                # Log problem
                LOGGER.error("Connection lost, reconnecting", exc_info=e)
                continue

            except Exception as e:
                LOGGER.critical(e, exc_info=True)

                channel.stop_consuming()
                break



from setuptools import find_packages, setup

with open("README.md") as readme_file:
    _long_description = readme_file.read()

setup(
    name="asset_scanner",
    version="0.2.2",
    description="Framework to provide plugin architecture to allow the scanning of assets to extract metadata and facets.",
    author="Richard Smith",
    url="https://github.com/cedadev/asset-scanner/",
    long_description=_long_description,
    long_description_content_type="text/markdown",
    license="BSD - See asset_extractor/LICENSE file for details",
    packages=find_packages(),
    test_suite="tests",
    package_data={"asset_scanner": ["LICENSE"]},
    install_requires=["pyyaml", "ceda-directory-tree", "tqdm"],
    extras_require={
        "docs": [
            "sphinx",
            "sphinx-rtd-theme",
            "elasticsearch",
            "sphinxcontrib-programoutput",
        ],
        "elasticsearch": ["elasticsearch"],
        "intake-esm": ["intake-esm"],
        "rabbitmq": ["pika"],
    },
    python_requires=">=3.5",
    tests_require=["pytest"],
    entry_points={
        "console_scripts": [
            "asset_scanner = asset_scanner.scripts.asset_scanner:main",
        ],
        "asset_scanner.output_plugins": [
            "standard_out = asset_scanner.plugins.output_plugins.standard_out:StdoutOutputBackend",
            "elasticsearch = asset_scanner.plugins.output_plugins.elasticsearch_backend:ElasticsearchOutputBackend",
        ],
        "asset_scanner.plugin_filters": [
            "path_regex = asset_scanner.plugins.filters.path_regex:PathRegexFilter",
        ],
        "asset_scanner.input_plugins": [
            "file_system = asset_scanner.plugins.input_plugins.file_system_input:FileSystemInputPlugin",
            "object_store = asset_scanner.plugins.input_plugins.object_store_input:ObjectStoreInputPlugin",
            "intake_esm = asset_scanner.plugins.input_plugins.intake_esm_input:IntakeESMInputPlugin",
            "rabbit_mq = asset_scanner.plugins.input_plugins.rabbit_mq_input:RabbitMQInputPlugin",
            "thredds = asset_scanner.plugins.input_plugins.thredds_input:ThreddsInputPlugin",
        ],
    },
)



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "01 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

import logging
from typing import Dict, List, Optional, Union

import pkg_resources

from .processor import BaseProcessor

LOGGER = logging.getLogger(__name__)


class HandlerPicker:
    """Loads the entrypoints

    Attributes:
        handlers - Dictionary of entry points. The dict is structured:

        .. code-block::

            {
                <entry-point-name>: <entry-point>,
                ...
            }
    """

    def __init__(self, entry_point_key: Union[List, str]):
        """
        Entry points to load from in the setup.py

        :param entry_point_key: name of the entry point source
        """
        self.handlers = {}

        if not entry_point_key:
            raise ValueError("No entry point specified. No handlers will be loaded")

        self.handlers = self._get_entrypoints(entry_point_key)

    @staticmethod
    def _get_entrypoints(group) -> Dict:
        """Get entrypoints for given group

        :param group: The named entry group
        :return: dict of entrypoints
        """
        entry_points = {}

        for entry_point in pkg_resources.iter_entry_points(group):
            entry_points[entry_point.name] = entry_point

        return entry_points

    def get_processor(self, name: str, **kwargs) -> Optional[BaseProcessor]:
        """
        Get the processor by name

        :param name: The name of the requested processor

        :return: Processor class
        """

        # Get the processor
        entry_point = self.handlers.get(name)

        if not entry_point:
            LOGGER.error(f"Failed to load processor: {name}")
            return

        # Try to load the processor
        processor = entry_point.load()

        return processor(**kwargs)



# encoding: utf-8
"""

"""
__author__ = "Richard Smith"
__date__ = "08 Jun 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

from .extractor import BaseExtractor

__all__ = ["BaseExtractor"]



# encoding: utf-8
"""
Intake Input
-----------------

Uses an `Intake catalog <https://intake.readthedocs.io/>`_
as a source for file objects.

**Plugin name:** ``intake_esm``

.. list-table::
    :header-rows: 1

    * - Option
      - Value Type
      - Description
    * - ``uri``
      - ``string``
      - ``REQUIRED`` The URI of a path or URL to an ESM collection JSON file.
    * - ``object_path_attr``
      - ``string``
      - ``REQUIRED`` The column header which contains the URI to
        the file object.
    * - ``catalog_kwargs``
      - ``dict``
      - Optional kwargs to pass to
        `intake.open_esm_datastore
        <https://intake-esm.readthedocs.io/en/latest
        /api.html#intake_esm.core.esm_datastore>`_
    * - ``search_kwargs``
      - ``dict``
      - Optional kwargs to pass to `esm_datastore.search
        <https://intake-esm.readthedocs.io/en/latest
        /api.html#intake_esm.core.esm_datastore.search>`_


Example Configuration:
    .. code-block:: yaml

        inputs:
            - name: intake_catalog
              uri: test_directory

"""
__author__ = "Richard Smith"
__date__ = "23 Sep 2021"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__contact__ = "richard.d.smith@stfc.ac.uk"

# Python imports
import logging
from datetime import datetime
from typing import TYPE_CHECKING
from urllib.parse import urlparse

# Thirdparty imports
import intake

# Framework imports
from asset_scanner.types.source_media import StorageType

# Package imports
from .base import BaseInputPlugin

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from asset_scanner.core import BaseExtractor


class IntakeESMInputPlugin(BaseInputPlugin):
    """
    Performs an os.walk to provide a stream of paths for procesing.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.uri = kwargs["uri"]
        self.object_attr = kwargs["object_path_attr"]

        self.intake_kwargs = kwargs.get("catalog_kwargs", {})
        self.search_kwargs = kwargs.get("search_kwargs")

    def open_catalog(self):
        """Open the ESM catalog and perform a search, if required."""
        logger.info("Opening catalog")
        catalog = intake.open_esm_datastore(self.uri, **self.intake_kwargs)

        if self.search_kwargs:
            catalog = catalog.search(**self.search_kwargs)

        logger.info(f"Found {len(catalog.df)} items")
        return catalog

    def run(self, extractor: "BaseExtractor"):
        total_files = 0
        start = datetime.now()

        catalog = self.open_catalog()

        for index, row in catalog.df.iterrows():
            filepath = getattr(row, self.object_attr)

            parse_result = urlparse(filepath)

            # Set media type
            media_type = StorageType.OBJECT_STORE
            if not parse_result.netloc:
                media_type = StorageType.POSIX

            if self.should_process(filepath, media_type):
                extractor.process_file(filepath, media_type)
                logger.debug(f"Input processing: {filepath}")
            else:
                logger.debug(f"Input skipping: {filepath}")

            total_files += 1

        end = datetime.now()
        print(f"Processed {total_files} files from {self.uri} in {end-start}")
